@@ -40,7 +40,6 @@
 #include "lm-internals.h"
 #include "lm-parser.h"
 #include "lm-sha.h"
-#include "lm-queue.h"
 #include "lm-connection.h"
 
 #define IN_BUFFER_SIZE 1024
@@ -59,6 +58,7 @@ struct _LmConnection {
 	/* Parameters */
 	GMainContext *context;
 	gchar        *server;
+	gchar        *jid;
 	guint         port;
 
 	LmSSL        *ssl;
@@ -88,11 +88,14 @@ struct _LmConnection {
 
 	LmCallback   *disconnect_cb;
 
-	LmQueue      *incoming_messages;
+	GQueue       *incoming_messages;
 	GSource      *incoming_source;
 
 	LmConnectionState state;
 
+	guint         keep_alive_rate;
+	guint         keep_alive_id;
+
 	gint          ref_count;
 };
 
@@ -166,8 +169,9 @@ static guint    connection_add_watch            (LmConnection  *connection,
 						 GIOCondition   condition,
 						 GIOFunc        func,
 						 gpointer       user_data);
-
-
+static gboolean connection_send_keep_alive      (LmConnection  *connection);
+static void     connection_start_keep_alive     (LmConnection  *connection);
+static void     connection_stop_keep_alive      (LmConnection  *connection);
 
 static GSourceFuncs incoming_funcs = {
 	connection_incoming_prepare,
@@ -182,6 +186,7 @@ connection_free (LmConnection *connection)
 	int i;
 
 	g_free (connection->server);
+	g_free (connection->jid);
 
 	/* Unref handlers */
 	for (i = 0; i < LM_MESSAGE_TYPE_UNKNOWN; ++i) {
@@ -259,15 +264,19 @@ connection_new_message_cb (LmParser     *parser,
 		    _lm_message_type_to_string (lm_message_get_type (m)),
 		    lm_message_node_get_attribute (m->node, "from"));
 
-	lm_queue_push_tail (connection->incoming_messages, m);
+	g_queue_push_tail (connection->incoming_messages, m);
 }
 
 gboolean
 _lm_connection_succeeded (LmConnectData *connect_data)
 {
-	LmConnection *connection = connect_data->connection;
+	LmConnection *connection;
 	LmMessage    *m;
 	GIOFlags      flags;
+	gchar        *server_from_jid;
+	gchar        *ch;
+
+	connection = connect_data->connection;
 
 	/* Need some way to report error/success */
 	if (connection->cancel_open) {
@@ -292,8 +301,9 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 	g_io_channel_set_flags (connection->io_channel, flags, NULL);
 
 	if (connection->ssl) {
-		if (!_lm_ssl_begin (connection->ssl, connection->fd, 
-				    connection->server, NULL)) {
+		if (!_lm_ssl_begin (connection->ssl, connection->fd,
+				    connection->server,
+				    NULL)) {
 			shutdown (connection->fd, SHUT_RDWR);
 			close (connection->fd);
 			connection_do_close (connection);
@@ -337,7 +347,13 @@ _lm_connection_succeeded (LmConnectData *connect_data)
 		return FALSE;
 	}
 
-	m = lm_message_new (connection->server, LM_MESSAGE_TYPE_STREAM);
+	if (connection->jid != NULL && (ch = strchr (connection->jid, '@')) != NULL) {
+		server_from_jid = ch + 1;
+	} else {
+		server_from_jid = connection->server;
+	}
+
+	m = lm_message_new (server_from_jid, LM_MESSAGE_TYPE_STREAM);
 	lm_message_node_set_attributes (m->node,
 					"xmlns:stream", 
 					"http://etherx.jabber.org/streams",
@@ -507,6 +523,41 @@ connection_add_watch (LmConnection *connection,
 	return id;
 }
 
+static gboolean
+connection_send_keep_alive (LmConnection *connection)
+{ 
+	if (!connection_send (connection, " ", -1, NULL)) {
+		lm_verbose ("Error while sending keep alive package");
+	}
+
+	return TRUE;
+}
+
+static void
+connection_start_keep_alive (LmConnection *connection)
+{
+	if (connection->keep_alive_id != 0) {
+		connection_stop_keep_alive (connection);
+	}
+
+	if (connection->keep_alive_rate > 0) {
+		connection->keep_alive_id =
+			g_timeout_add (connection->keep_alive_rate,
+				       (GSourceFunc) connection_send_keep_alive,
+				       connection);
+	}
+}
+
+static void
+connection_stop_keep_alive (LmConnection *connection)
+{
+	if (connection->keep_alive_id != 0) {
+		g_source_remove (connection->keep_alive_id);
+	}
+
+	connection->keep_alive_id = 0;
+}
+
 /* Returns directly */
 /* Setups all data needed to start the connection attempts */
 static gboolean
@@ -546,7 +597,7 @@ connection_do_open (LmConnection *connection, GError **error)
 	req.ai_protocol = IPPROTO_TCP;
 	
 	connection->cancel_open = FALSE;
-	connection->state = LM_CONNECTION_STATE_CONNECTING;
+	connection->state = LM_CONNECTION_STATE_OPENING;
 	
 	if (connection->proxy) {
 		const gchar *proxy_server;
@@ -565,9 +616,11 @@ connection_do_open (LmConnection *connection, GError **error)
 		}
 	} else { /* connect directly */
 		g_log (LM_LOG_DOMAIN,LM_LOG_LEVEL_NET,
-		       "Going to connect to %s\n",connection->server);
+		       "Going to connect to %s\n", 
+		       connection->server);
 
-		if (getaddrinfo (connection->server, NULL, &req, &ans) != 0) {
+		if (getaddrinfo (connection->server,
+				 NULL, &req, &ans) != 0) {
 			g_set_error (error,
 				     LM_ERROR,                 
 				     LM_ERROR_CONNECTION_OPEN,   
@@ -596,6 +649,8 @@ connection_do_open (LmConnection *connection, GError **error)
 static void
 connection_do_close (LmConnection *connection)
 {
+	connection_stop_keep_alive (connection);
+
 	if (connection->io_channel) {
 		g_source_remove (connection->io_watch_in);
 		g_source_remove (connection->io_watch_err);
@@ -611,8 +666,8 @@ connection_do_close (LmConnection *connection)
 	if (!lm_connection_is_open (connection)) {
 		return;
 	}
-
-	connection->state = LM_CONNECTION_STATE_DISCONNECTED;
+	
+	connection->state = LM_CONNECTION_STATE_CLOSED;
 
 	if (connection->ssl) {
 		_lm_ssl_close (connection->ssl);
@@ -724,7 +779,7 @@ connection_send (LmConnection  *connection,
 {
 	gsize             bytes_written;
 	
-	if (connection->state < LM_CONNECTION_STATE_CONNECTING) {
+	if (connection->state < LM_CONNECTION_STATE_OPENING) {
 		g_set_error (error,
 			     LM_ERROR,
 			     LM_ERROR_CONNECTION_NOT_OPEN,
@@ -915,7 +970,7 @@ connection_auth_reply (LmMessageHandler *handler,
 	} 
 	else if (strcmp (type, "error") == 0) {
 		result = FALSE;
-		connection->state = LM_CONNECTION_STATE_CONNECTED;
+		connection->state = LM_CONNECTION_STATE_OPEN;
 	}
 	
 	lm_verbose ("AUTH reply: %d\n", result);
@@ -947,11 +1002,13 @@ connection_stream_received (LmConnection *connection, LmMessage *m)
 	
 	lm_verbose ("Stream received: %s\n", connection->stream_id);
 	
-	connection->state = LM_CONNECTION_STATE_CONNECTED;
+	connection->state = LM_CONNECTION_STATE_OPEN;
 	
 	/* Check to see if the stream is correctly set up */
 	result = TRUE;
 
+	connection_start_keep_alive (connection);
+
 	if (connection->open_cb && connection->open_cb->func) {
 		LmCallback *cb = connection->open_cb;
 		
@@ -976,7 +1033,7 @@ connection_incoming_prepare (GSource *source, gint *timeout)
 	
 	connection = ((LmIncomingSource *)source)->connection;
 	
-	return !lm_queue_is_empty (connection->incoming_messages);
+	return !g_queue_is_empty (connection->incoming_messages);
 }
 
 static gboolean
@@ -995,7 +1052,7 @@ connection_incoming_dispatch (GSource *source,
 	
 	connection = ((LmIncomingSource *) source)->connection;
 
-	m = (LmMessage *) lm_queue_pop_head (connection->incoming_messages);
+	m = (LmMessage *) g_queue_pop_head (connection->incoming_messages);
 	
 	if (m) {
 		connection_handle_message (connection, m);
@@ -1056,12 +1113,15 @@ lm_connection_new (const gchar *server)
 
 	connection->context           = NULL;
 	connection->port              = LM_CONNECTION_DEFAULT_PORT;
+	connection->jid               = NULL;
 	connection->ssl               = NULL;
 	connection->proxy             = NULL;
 	connection->disconnect_cb     = NULL;
-	connection->incoming_messages = lm_queue_new ();
+	connection->incoming_messages = g_queue_new ();
 	connection->cancel_open       = FALSE;
-	connection->state             = LM_CONNECTION_STATE_DISCONNECTED;
+	connection->state             = LM_CONNECTION_STATE_CLOSED;
+	connection->keep_alive_id     = 0;
+	connection->keep_alive_rate   = 0;
 	
 	connection->id_handlers = g_hash_table_new_full (g_str_hash, 
 							 g_str_equal,
@@ -1155,7 +1215,7 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 		return FALSE;
 	}
 	
-	while ((state = lm_connection_get_state (connection)) == LM_CONNECTION_STATE_CONNECTING) {
+	while ((state = lm_connection_get_state (connection)) == LM_CONNECTION_STATE_OPENING) {
 		if (g_main_context_pending (connection->context)) {
 			g_main_context_iteration (connection->context, TRUE);
 		} else {
@@ -1164,6 +1224,7 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
 	}
 
 	if (lm_connection_is_open (connection)) {
+		connection_start_keep_alive (connection);
 		return TRUE;
 	}
 
@@ -1174,7 +1235,7 @@ lm_connection_open_and_block (LmConnection *connection, GError **error)
  * lm_connection_cancel_open:
  * @connection: an #LmConnection to cancel opening on
  *
- * Cancels the open operation of a connection. The connection should be in the state #LM_CONNECTION_STATE_CONNECTING.
+ * Cancels the open operation of a connection. The connection should be in the state #LM_CONNECTION_STATE_OPENING.
  **/
 void
 lm_connection_cancel_open (LmConnection *connection)
@@ -1206,10 +1267,9 @@ lm_connection_close (LmConnection      *connection,
 			     "Connection is not open, call lm_connection_open() first");
 		return FALSE;
 	}
-	
+
 	lm_verbose ("Disconnecting from: %s:%d\n", 
-		    connection->server,
-		    connection->port);
+		    connection->server, connection->port);
 	
 	if (!connection_send (connection, "</stream:stream>", -1, error)) {
 		return FALSE;
@@ -1368,6 +1428,33 @@ lm_connection_authenticate_and_block (LmConnection  *connection,
 	return FALSE;
 }
 
+/**
+ * lm_connection_set_keep_alive_rate:
+ * @connection: #LmConnection to check if it is open.
+ * @rate: Number of seconds between keep alive packages are sent.
+ * 
+ * Set the keep alive rate, in seconds. Set to 0 to prevent keep alive messages to be sent.
+ * A keep alive message is a single space character.
+ **/
+void
+lm_connection_set_keep_alive_rate (LmConnection *connection, guint rate)
+{
+	g_return_if_fail (connection != NULL);
+
+	connection_stop_keep_alive (connection);
+
+	if (rate == 0) {
+		connection->keep_alive_id = 0;
+		return;
+	}
+
+	connection->keep_alive_rate = rate * 1000;
+	
+	if (lm_connection_is_open (connection)) {
+		connection_start_keep_alive (connection);
+	}
+}
+
 /**
  * lm_connection_is_open:
  * @connection: #LmConnection to check if it is open.
@@ -1381,7 +1468,7 @@ lm_connection_is_open (LmConnection *connection)
 {
 	g_return_val_if_fail (connection != NULL, FALSE);
 	
-	return connection->state >= LM_CONNECTION_STATE_CONNECTED;
+	return connection->state >= LM_CONNECTION_STATE_OPEN;
 }
 
 /**
@@ -1404,7 +1491,7 @@ lm_connection_is_authenticated (LmConnection *connection)
  * lm_connection_get_server:
  * @connection: an #LmConnection
  * 
- * Fetches the server address that @connection is using.
+ * Fetches the server address that @connection is using. 
  * 
  * Return value: the server address
  **/
@@ -1421,7 +1508,8 @@ lm_connection_get_server (LmConnection *connection)
  * @connection: an #LmConnection
  * @server: Address of the server
  * 
- * Sets the server address for @connection to @server. Notice that @connection can't be open while doing this.
+ * Sets the server address for @connection to @server. Notice that @connection
+ * can't be open while doing this. 
  **/
 void
 lm_connection_set_server (LmConnection *connection, const gchar *server)
@@ -1434,13 +1522,47 @@ lm_connection_set_server (LmConnection *connection, const gchar *server)
 		return;
 	}
 	
-	if (connection->server) {
-		g_free (connection->server);
-	}
-	
+	g_free (connection->server);
 	connection->server = g_strdup (server);
 }
 
+/**
+ * lm_connection_get_jid:
+ * @connection: an #LmConnection
+ * 
+ * Fetches the jid set for @connection is using. 
+ * 
+ * Return value: the jid
+ **/
+const gchar *
+lm_connection_get_jid (LmConnection *connection)
+{
+	g_return_val_if_fail (connection != NULL, NULL);
+
+	return connection->jid;
+}
+
+/**
+ * lm_connection_set_jid:
+ * @connection: an #LmConnection
+ * @jid: JID to be used for @connection
+ * 
+ * Sets the JID to be used for @connection.
+ **/
+void 
+lm_connection_set_jid (LmConnection *connection, const gchar *jid)
+{
+	g_return_if_fail (connection != NULL);
+
+	if (lm_connection_is_open (connection)) {
+		g_warning ("Can't change JID while connected");
+		return;
+	}
+
+	g_free (connection->jid);
+	connection->jid = g_strdup (jid);
+}
+
 /**
  * lm_connection_get_port:
  * @connection: an #LmConnection
@@ -1537,7 +1659,8 @@ lm_connection_get_proxy (LmConnection *connection)
  * @connection: an #LmConnection
  * @proxy: an #LmProxy
  *
- * Sets the proxy to use for this connection.
+ * Sets the proxy to use for this connection. To unset pass a proxy of type 
+ * #LM_PROXY_TYPE_NONE to this function.
  * 
  **/
 void
@@ -1550,12 +1673,15 @@ lm_connection_set_proxy (LmConnection *connection, LmProxy *proxy)
 		g_warning ("Can't change server proxy while connected");
 		return;
 	}
-	
+
 	if (connection->proxy) {
 		lm_proxy_unref (connection->proxy);
+		connection->proxy = NULL;
 	}
 
-	connection->proxy = lm_proxy_ref (proxy);
+	if (lm_proxy_get_type (proxy) != LM_PROXY_TYPE_NONE) {
+		connection->proxy = lm_proxy_ref (proxy);
+	}
 }
 
 /**
@@ -1668,20 +1794,21 @@ lm_connection_send_with_reply_and_block (LmConnection  *connection,
 
 		g_main_context_iteration (connection->context, TRUE);
 	
-		if (lm_queue_is_empty (connection->incoming_messages)) {
+		if (g_queue_is_empty (connection->incoming_messages)) {
 			continue;
 		}
 
-		for (n = 0; n < connection->incoming_messages->length; n++) {
+		for (n = 0; n < g_queue_get_length (connection->incoming_messages); n++) {
 			LmMessage *m;
 
-			m = lm_queue_peek_nth (connection->incoming_messages, n);
+			m = (LmMessage *) g_queue_peek_nth (connection->incoming_messages, n);
 
 			m_id = lm_message_node_get_attribute (m->node, "id");
 			
 			if (m_id && strcmp (m_id, id) == 0) {
 				reply = m;
-				lm_queue_remove_nth (connection->incoming_messages, n);
+				g_queue_pop_nth (connection->incoming_messages,
+						 n);
 				break;
 			}
 		}
@@ -1822,7 +1949,7 @@ LmConnectionState
 lm_connection_get_state (LmConnection *connection)
 {
 	g_return_val_if_fail (connection != NULL, 
-			      LM_CONNECTION_STATE_DISCONNECTED);
+			      LM_CONNECTION_STATE_CLOSED);
 
 	return connection->state;
 }
